"""AnyDI Typer extension."""

from __future__ import annotations

import contextlib
import functools
import inspect
from collections.abc import Awaitable, Callable
from typing import Any

import anyio
from typer import Typer

from anydi import Container, Scope

__all__ = ["install"]


def _wrap_async_callback_no_injection(callback: Callable[..., Any]) -> Any:
    """Wrap async callback without injection in anyio.run()."""

    @functools.wraps(callback)
    def async_no_injection_wrapper(*args: Any, **kwargs: Any) -> Any:
        return anyio.run(callback, *args, **kwargs)

    return async_no_injection_wrapper


def _wrap_async_callback_with_injection(
    callback: Callable[..., Awaitable[Any]],
    container: Container,
    sig: inspect.Signature,
    non_injected_params: set[inspect.Parameter],
    scopes: set[Scope],
) -> Any:
    """Wrap async callback with injection in anyio.run()."""

    @functools.wraps(callback)
    def async_wrapper(*args: Any, **kwargs: Any) -> Any:
        async def _run() -> Any:
            ordered_scopes = container.get_ordered_scopes(scopes)

            async with contextlib.AsyncExitStack() as stack:
                # Start scoped contexts in dependency order
                for scope in ordered_scopes:
                    if scope == "singleton":
                        await stack.enter_async_context(container)
                    else:
                        await stack.enter_async_context(
                            container.ascoped_context(scope)
                        )

                return await container.run(callback, *args, **kwargs)

        return anyio.run(_run)

    # Update the wrapper's signature to only show non-injected parameters to Typer
    async_wrapper.__signature__ = sig.replace(parameters=non_injected_params)  # type: ignore

    return async_wrapper


def _process_callback(callback: Callable[..., Any], container: Container) -> Any:
    """Validate and wrap a callback for dependency injection."""
    sig = inspect.signature(callback, eval_str=True)
    injected_param_names: set[str] = set()
    non_injected_params: set[inspect.Parameter] = set()
    scopes: set[Scope] = set()

    # Validate parameters and collect which ones need injection
    for parameter in sig.parameters.values():
        interface, should_inject = container.validate_injected_parameter(
            parameter, call=callback
        )
        if should_inject:
            injected_param_names.add(parameter.name)
            scopes.add(container.providers[interface].scope)
        else:
            non_injected_params.add(parameter)

    # If no parameters need injection and callback is not async, return original
    if not injected_param_names and not inspect.iscoroutinefunction(callback):
        return callback

    # If async callback with no injection, just wrap in anyio.run()
    if not injected_param_names and inspect.iscoroutinefunction(callback):
        return _wrap_async_callback_no_injection(callback)

    # Handle async callbacks - wrap them in anyio.run() for Typer
    if inspect.iscoroutinefunction(callback):
        return _wrap_async_callback_with_injection(
            callback, container, sig, non_injected_params, scopes
        )

    @functools.wraps(callback)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        ordered_scopes = container.get_ordered_scopes(scopes)

        with contextlib.ExitStack() as stack:
            # Start scoped contexts in dependency order
            for scope in ordered_scopes:
                if scope == "singleton":
                    stack.enter_context(container)
                else:
                    stack.enter_context(container.scoped_context(scope))

            return container.run(callback, *args, **kwargs)

    # Update the wrapper's signature to only show non-injected parameters to Typer
    wrapper.__signature__ = sig.replace(parameters=non_injected_params)  # type: ignore

    return wrapper


def install(app: Typer, container: Container) -> None:
    """Install AnyDI into a Typer application."""
    # Process main callback if exists
    if app.registered_callback:
        callback = app.registered_callback.callback
        if callback:
            app.registered_callback.callback = _process_callback(callback, container)

    # Process all registered commands
    for command_info in app.registered_commands:
        callback = command_info.callback
        if callback:
            command_info.callback = _process_callback(callback, container)

    # Process nested Typer groups
    for group_info in app.registered_groups:
        # Recursively install for nested Typer apps
        if group_info.typer_instance:
            install(group_info.typer_instance, container)



"""Shared AnyDI utils module."""

from __future__ import annotations

import inspect
from collections.abc import AsyncIterator, Callable, Iterator
from types import NoneType
from typing import TYPE_CHECKING, Annotated, Any, Literal, TypeVar

from typing_extensions import Sentinel

T = TypeVar("T")

Scope = Literal["transient", "singleton", "request"] | str

NOT_SET = Sentinel("NOT_SET")


class ProvideMarker:
    """A marker object for declaring dependency."""

    __slots__ = ("_interface",)

    def __init__(self, interface: Any = NOT_SET) -> None:
        self._interface = interface

    @property
    def interface(self) -> Any:
        if self._interface is NOT_SET:
            raise TypeError("Interface is not set.")
        return self._interface

    @interface.setter
    def interface(self, interface: Any) -> None:
        self._interface = interface

    def __class_getitem__(cls, item: Any) -> Any:
        return Annotated[item, cls()]


_provide_factory: Callable[[], Any] = ProvideMarker


def set_provide_factory(factory: Callable[[], Any]) -> Callable[[], Any]:
    """Set the global factory used by Inject() and Provide."""
    global _provide_factory
    previous = _provide_factory
    _provide_factory = factory
    return previous


def is_provide_marker(obj: Any) -> bool:
    return isinstance(obj, ProvideMarker)


class _ProvideMeta(type):
    """Metaclass for Provide that delegates __class_getitem__ to the current factory."""

    def __getitem__(cls, item: Any) -> Any:
        # Use the current factory's __class_getitem__ if available
        factory = _provide_factory
        if hasattr(factory, "__class_getitem__"):
            return factory.__class_getitem__(item)  # type: ignore[attr-defined]
        # Fallback to creating Annotated with factory instance
        return Annotated[item, factory()]


if TYPE_CHECKING:
    Provide = Annotated[T, ProvideMarker()]

else:

    class Provide(metaclass=_ProvideMeta):
        pass


def Inject() -> Any:
    return _provide_factory()


# Alias from backward compatibility
is_inject_marker = is_provide_marker


class Event:
    """Represents an event object."""

    __slots__ = ()


def is_event_type(obj: Any) -> bool:
    """Checks if an object is an event type."""
    return inspect.isclass(obj) and issubclass(obj, Event)


def is_context_manager(obj: Any) -> bool:
    """Check if the given object is a context manager."""
    return hasattr(obj, "__enter__") and hasattr(obj, "__exit__")


def is_async_context_manager(obj: Any) -> bool:
    """Check if the given object is an async context manager."""
    return hasattr(obj, "__aenter__") and hasattr(obj, "__aexit__")


def is_none_type(tp: Any) -> bool:
    """Check if the given object is a None type."""
    return tp in (None, NoneType)


def is_iterator_type(tp: Any) -> bool:
    """Check if the given object is an iterator type."""
    return tp in (Iterator, AsyncIterator)
